@@ -32,6 +32,7 @@
 #include <osmocom/core/timer.h>
 #include <osmocom/core/msgb.h>
 #include <osmocom/core/talloc.h>
+#include <osmocom/core/tdef.h>
 #include <osmocom/core/stats.h>
 #include <osmocom/core/rate_ctr.h>
 #include <osmocom/gsm/protocol/gsm_12_21.h>
@@ -50,6 +51,9 @@
 #include <osmo-bts/signal.h>
 #include <osmo-bts/dtx_dl_amr_fsm.h>
 #include <osmo-bts/cbch.h>
+#include <osmo-bts/bts_shutdown_fsm.h>
+#include <osmo-bts/nm_common_fsm.h>
+#include <osmo-bts/power_control.h>
 
 #define MIN_QUAL_RACH	 50 /* minimum link quality (in centiBels) for Access Bursts */
 #define MIN_QUAL_NORM	 -5 /* minimum link quality (in centiBels) for Normal Bursts */
@@ -122,6 +126,172 @@ static const struct rate_ctr_group_desc cbch_ctrg_desc = {
 	cbch_ctr_desc
 };
 
+static struct osmo_tdef bts_T_defs[] = {
+	/* T-1: FIXME: Ideally should be dynamically calculated per trx at
+	 * shutdown start based on params below, and highest trx value taken:
+	 * + VTY's power-ramp step-interval.
+	 * + Amount of steps needed (taking into account how many dB each step moves).
+	 * + Extra time to get response back for each step.
+	 * For now we simply give 5 mins, which should be enough for any
+	 * acceptable setup, while still ensuring will timeout at some point if
+	 * something fails in the ramp down procedure.
+	 */
+	{ .T=-1, .default_val=300, .desc="Time after which osmo-bts exits if regular ramp down during shut down process does not finish (s)" },
+	{ .T=-2, .default_val=3, .desc="Time after which osmo-bts exits if requesting transceivers to stop during shut down process does not finish (s)" },
+	{}
+};
+
+static const uint8_t bts_nse_timer_default[] = { 3, 3, 3, 3, 30, 3, 10 };
+static const uint8_t bts_cell_timer_default[] =
+				{ 3, 3, 3, 3, 3, 10, 3, 10, 3, 10, 3 };
+static const struct gprs_rlc_cfg rlc_cfg_default = {
+	.parameter = {
+		[RLC_T3142] = 20,
+		[RLC_T3169] = 5,
+		[RLC_T3191] = 5,
+		[RLC_T3193] = 160, /* 10ms */
+		[RLC_T3195] = 5,
+		[RLC_N3101] = 10,
+		[RLC_N3103] = 4,
+		[RLC_N3105] = 8,
+		[CV_COUNTDOWN] = 15,
+		[T_DL_TBF_EXT] = 250 * 10, /* ms */
+		[T_UL_TBF_EXT] = 250 * 10, /* ms */
+	},
+	.paging = {
+		.repeat_time = 5 * 50, /* ms */
+		.repeat_count = 3,
+	},
+	.cs_mask = 0x1fff,
+	.initial_cs = 2,
+	.initial_mcs = 6,
+};
+
+const struct value_string osmo_bts_variant_names[_NUM_BTS_VARIANT + 1] = {
+	{ BTS_UNKNOWN,		"unknown" },
+	{ BTS_OSMO_LITECELL15,	"osmo-bts-lc15" },
+	{ BTS_OSMO_OC2G,	"osmo-bts-oc2g" },
+	{ BTS_OSMO_OCTPHY,	"osmo-bts-octphy" },
+	{ BTS_OSMO_SYSMO,	"osmo-bts-sysmo" },
+	{ BTS_OSMO_TRX,		"osmo-bts-trx" },
+	{ BTS_OSMO_VIRTUAL,	"osmo-bts-virtual" },
+	{ BTS_OSMO_OMLDUMMY,	"osmo-bts-omldummy" },
+	{ 0, NULL }
+};
+
+const char *btsvariant2str(enum gsm_bts_type_variant v)
+{
+	return get_value_string(osmo_bts_variant_names, v);
+}
+
+const struct value_string bts_attribute_names[] = {
+	OSMO_VALUE_STRING(BTS_TYPE_VARIANT),
+	OSMO_VALUE_STRING(BTS_SUB_MODEL),
+	OSMO_VALUE_STRING(TRX_PHY_VERSION),
+	{ 0, NULL }
+};
+
+const char *btsatttr2str(enum bts_attribute v)
+{
+	return get_value_string(bts_attribute_names, v);
+}
+
+const struct value_string bts_impl_flag_desc[] = {
+	{ BTS_INTERNAL_FLAG_MS_PWR_CTRL_DSP,	"DSP/HW based MS Power Control Loop" },
+	{ BTS_INTERNAL_FLAG_MEAS_PAYLOAD_COMB,	"Measurement and Payload data combined" },
+	{ 0, NULL }
+};
+
+static int gsm_bts_talloc_destructor(struct gsm_bts *bts)
+{
+	if (bts->site_mgr.mo.fi) {
+		osmo_fsm_inst_free(bts->site_mgr.mo.fi);
+		bts->site_mgr.mo.fi = NULL;
+	}
+	if (bts->mo.fi) {
+		osmo_fsm_inst_free(bts->mo.fi);
+		bts->mo.fi = NULL;
+	}
+	return 0;
+}
+
+struct gsm_bts *gsm_bts_alloc(void *ctx, uint8_t bts_num)
+{
+	struct gsm_bts *bts = talloc_zero(ctx, struct gsm_bts);
+	int i;
+
+	if (!bts)
+		return NULL;
+
+	talloc_set_destructor(bts, gsm_bts_talloc_destructor);
+
+	bts->nr = bts_num;
+	bts->num_trx = 0;
+	INIT_LLIST_HEAD(&bts->trx_list);
+	bts->ms_max_power = 15;	/* dBm */
+
+	bts->T_defs = bts_T_defs;
+	osmo_tdefs_reset(bts->T_defs);
+	bts->shutdown_fi = osmo_fsm_inst_alloc(&bts_shutdown_fsm, bts, bts,
+					       LOGL_INFO, NULL);
+	osmo_fsm_inst_update_id_f(bts->shutdown_fi, "bts%d", bts->nr);
+
+	bts->site_mgr.mo.fi = osmo_fsm_inst_alloc(&nm_bts_sm_fsm, bts, &bts->site_mgr,
+						  LOGL_INFO, "bts_sm");
+	gsm_mo_init(&bts->site_mgr.mo, bts, NM_OC_SITE_MANAGER,
+		    0xff, 0xff, 0xff);
+
+	bts->mo.fi = osmo_fsm_inst_alloc(&nm_bts_fsm, bts, bts,
+					 LOGL_INFO, NULL);
+	osmo_fsm_inst_update_id_f(bts->mo.fi, "bts%d", bts->nr);
+	gsm_mo_init(&bts->mo, bts, NM_OC_BTS, bts->nr, 0xff, 0xff);
+
+	for (i = 0; i < ARRAY_SIZE(bts->gprs.nsvc); i++) {
+		bts->gprs.nsvc[i].bts = bts;
+		bts->gprs.nsvc[i].id = i;
+		gsm_mo_init(&bts->gprs.nsvc[i].mo, bts, NM_OC_GPRS_NSVC,
+				bts->nr, i, 0xff);
+	}
+	memcpy(&bts->gprs.nse.timer, bts_nse_timer_default,
+		sizeof(bts->gprs.nse.timer));
+	gsm_mo_init(&bts->gprs.nse.mo, bts, NM_OC_GPRS_NSE,
+			bts->nr, 0xff, 0xff);
+	memcpy(&bts->gprs.cell.timer, bts_cell_timer_default,
+		sizeof(bts->gprs.cell.timer));
+	gsm_mo_init(&bts->gprs.cell.mo, bts, NM_OC_GPRS_CELL,
+			bts->nr, 0xff, 0xff);
+	memcpy(&bts->gprs.cell.rlc_cfg, &rlc_cfg_default,
+		sizeof(bts->gprs.cell.rlc_cfg));
+
+	/* create our primary TRX. It will be initialized during bts_init() */
+	bts->c0 = gsm_bts_trx_alloc(bts);
+	if (!bts->c0) {
+		talloc_free(bts);
+		return NULL;
+	}
+	bts->c0->ts[0].pchan = GSM_PCHAN_CCCH_SDCCH4;
+
+	bts->features = bitvec_alloc(MAX_BTS_FEATURES / 8, bts);
+	OSMO_ASSERT(bts->features != NULL);
+
+	return bts;
+}
+
+struct gsm_bts *gsm_bts_num(const struct gsm_network *net, int num)
+{
+	struct gsm_bts *bts;
+
+	if (num >= net->num_bts)
+		return NULL;
+
+	llist_for_each_entry(bts, &net->bts_list, list) {
+		if (bts->nr == num)
+			return bts;
+	}
+
+	return NULL;
+}
+
 /* Initialize the BTS data structures, called before config
  * file reading */
 int bts_init(struct gsm_bts *bts)
@@ -154,11 +324,15 @@ int bts_init(struct gsm_bts *bts)
 
 	/* configurable via VTY */
 	bts->paging_state = paging_init(bts, 200, 0);
-	bts->ul_power_target = -75;	/* dBm default */
 	bts->rtp_jitter_adaptive = false;
 	bts->rtp_port_range_start = 16384;
 	bts->rtp_port_range_end = 17407;
 	bts->rtp_port_range_next = bts->rtp_port_range_start;
+	bts->rtp_ip_dscp = -1;
+
+	/* Default (fall-back) MS/BS Power control parameters */
+	bts->bs_dpc_params = power_ctrl_params_def;
+	bts->ms_dpc_params = power_ctrl_params_def;
 
 	/* configurable via OML */
 	bts->load.ccch.load_ind_period = 112;
@@ -175,16 +349,17 @@ int bts_init(struct gsm_bts *bts)
 		bts->t200_ms[i] = oml_default_t200_ms[i];
 
 	/* default RADIO_LINK_TIMEOUT */
-	bts->radio_link_timeout = 32;
+	bts->radio_link_timeout.oml = 32;
+	bts->radio_link_timeout.current = bts->radio_link_timeout.oml;
 
 	/* Start with the site manager */
-	oml_mo_state_init(&bts->site_mgr.mo, NM_OPSTATE_ENABLED, NM_AVSTATE_OK);
+	oml_mo_state_init(&bts->site_mgr.mo, NM_OPSTATE_DISABLED, NM_AVSTATE_NOT_INSTALLED);
+	oml_mo_state_init(&bts->mo, NM_OPSTATE_DISABLED, NM_AVSTATE_NOT_INSTALLED);
 
-	/* set BTS to dependency */
-	oml_mo_state_init(&bts->mo, -1, NM_AVSTATE_DEPENDENCY);
-	oml_mo_state_init(&bts->gprs.nse.mo, -1, NM_AVSTATE_DEPENDENCY);
-	oml_mo_state_init(&bts->gprs.cell.mo, -1, NM_AVSTATE_DEPENDENCY);
-	oml_mo_state_init(&bts->gprs.nsvc[0].mo, -1, NM_AVSTATE_DEPENDENCY);
+	/* set BTS attr to dependency */
+	oml_mo_state_init(&bts->gprs.nse.mo, NM_OPSTATE_DISABLED, NM_AVSTATE_DEPENDENCY);
+	oml_mo_state_init(&bts->gprs.cell.mo, NM_OPSTATE_DISABLED, NM_AVSTATE_DEPENDENCY);
+	oml_mo_state_init(&bts->gprs.nsvc[0].mo, NM_OPSTATE_DISABLED, NM_AVSTATE_DEPENDENCY);
 	oml_mo_state_init(&bts->gprs.nsvc[1].mo, NM_OPSTATE_DISABLED, NM_AVSTATE_OFF_LINE);
 
 	/* allocate a talloc pool for ORTP to ensure it doesn't have to go back
@@ -193,7 +368,8 @@ int bts_init(struct gsm_bts *bts)
 	osmo_rtp_init(tall_rtp_ctx);
 
 	/* features implemented in 'common', available for all models */
-	gsm_bts_set_feature(bts, BTS_FEAT_ETWS_PN);
+	osmo_bts_set_feature(bts->features, BTS_FEAT_ETWS_PN);
+	osmo_bts_set_feature(bts->features, BTS_FEAT_IPV6_NSVC);
 
 	rc = bts_model_init(bts);
 	if (rc < 0) {
@@ -234,83 +410,16 @@ int bts_init(struct gsm_bts *bts)
 	return rc;
 }
 
-/* Initialize the TRX data structures, called before config
- * file reading */
-int bts_trx_init(struct gsm_bts_trx *trx)
-{
-	/* initialize bts data structure */
-	struct trx_power_params *tpp = &trx->power_params;
-	int rc, i;
-
-	for (i = 0; i < ARRAY_SIZE(trx->ts); i++) {
-		struct gsm_bts_trx_ts *ts = &trx->ts[i];
-		int k;
-
-		for (k = 0; k < ARRAY_SIZE(ts->lchan); k++) {
-			struct gsm_lchan *lchan = &ts->lchan[k];
-			INIT_LLIST_HEAD(&lchan->dl_tch_queue);
-		}
-	}
-	/* Default values for the power adjustments */
-	tpp->ramp.max_initial_pout_mdBm = to_mdB(0);
-	tpp->ramp.step_size_mdB = to_mdB(2);
-	tpp->ramp.step_interval_sec = 1;
-
-	/* IF BTS model doesn't DSP/HW support MS Power Control Loop, enable osmo algo by default: */
-	if (!gsm_bts_has_feature(trx->bts, BTS_FEAT_MS_PWR_CTRL_DSP))
-		trx->ms_pwr_ctl_soft = true;
-
-	rc = bts_model_trx_init(trx);
-	if (rc < 0) {
-		llist_del(&trx->list);
-		return rc;
-	}
-	return 0;
-}
-
-static void shutdown_timer_cb(void *data)
-{
-	fprintf(stderr, "Shutdown timer expired\n");
-	exit(42);
-}
-
-static struct osmo_timer_list shutdown_timer = {
-	.cb = &shutdown_timer_cb,
-};
-
-void bts_shutdown(struct gsm_bts *bts, const char *reason)
-{
-	struct gsm_bts_trx *trx;
-
-	if (osmo_timer_pending(&shutdown_timer)) {
-		LOGP(DOML, LOGL_NOTICE,
-			"BTS is already being shutdown.\n");
-		return;
-	}
-
-	LOGP(DOML, LOGL_NOTICE, "Shutting down BTS %u, Reason %s\n",
-		bts->nr, reason);
-
-	llist_for_each_entry_reverse(trx, &bts->trx_list, list) {
-		bts_model_trx_deact_rf(trx);
-		bts_model_trx_close(trx);
-	}
-
-	/* schedule a timer to make sure select loop logic can run again
-	 * to dispatch any pending primitives */
-	osmo_timer_schedule(&shutdown_timer, 3, 0);
-}
-
 /* main link is established, send status report */
 int bts_link_estab(struct gsm_bts *bts)
 {
 	int i, j;
 
-	LOGP(DSUM, LOGL_INFO, "Main link established, sending Status'.\n");
+	LOGP(DSUM, LOGL_INFO, "Main link established, sending NM Status.\n");
 
-	/* BTS and SITE MGR are EANBLED, BTS is DEPENDENCY */
-	oml_tx_state_changed(&bts->site_mgr.mo);
-	oml_tx_state_changed(&bts->mo);
+	/* BTS SITE MGR becomes Offline (tx SW ACT Report), BTS is DEPENDENCY */
+	osmo_fsm_inst_dispatch(bts->site_mgr.mo.fi, NM_EV_SW_ACT, NULL);
+	osmo_fsm_inst_dispatch(bts->mo.fi, NM_EV_SW_ACT, NULL);
 
 	/* those should all be in DEPENDENCY */
 	oml_tx_state_changed(&bts->gprs.nse.mo);
@@ -335,55 +444,6 @@ int bts_link_estab(struct gsm_bts *bts)
 	return bts_model_oml_estab(bts);
 }
 
-/* RSL link is established, send status report */
-int trx_link_estab(struct gsm_bts_trx *trx)
-{
-	struct e1inp_sign_link *link = trx->rsl_link;
-	uint8_t radio_state = link ?  NM_OPSTATE_ENABLED : NM_OPSTATE_DISABLED;
-	int rc;
-
-	LOGP(DSUM, LOGL_INFO, "RSL link (TRX %02x) state changed to %s, sending Status'.\n",
-		trx->nr, link ? "up" : "down");
-
-	oml_mo_state_chg(&trx->mo, radio_state, NM_AVSTATE_OK);
-
-	if (link)
-		rc = rsl_tx_rf_res(trx);
-	else
-		rc = bts_model_trx_deact_rf(trx);
-	if (rc < 0) {
-		oml_tx_failure_event_rep(&trx->bb_transc.mo, NM_SEVER_MAJOR, OSMO_EVT_MAJ_RSL_FAIL,
-					 link ?
-					 "Failed to establish RSL link (%d)" :
-					 "Failed to deactivate RF (%d)", rc);
-	}
-
-	return 0;
-}
-
-/* set the availability of the TRX (used by PHY driver) */
-int trx_set_available(struct gsm_bts_trx *trx, int avail)
-{
-	int tn;
-
-	LOGP(DSUM, LOGL_INFO, "TRX(%d): Setting available = %d\n",
-		trx->nr, avail);
-	if (avail) {
-		int op_state = trx->rsl_link ?  NM_OPSTATE_ENABLED : NM_OPSTATE_DISABLED;
-		oml_mo_state_chg(&trx->mo, op_state, NM_AVSTATE_OK);
-		oml_mo_state_chg(&trx->bb_transc.mo, -1, NM_AVSTATE_OK);
-		for (tn = 0; tn < ARRAY_SIZE(trx->ts); tn++)
-			oml_mo_state_chg(&trx->ts[tn].mo, op_state, NM_AVSTATE_OK);
-	} else {
-		oml_mo_state_chg(&trx->mo,  NM_OPSTATE_DISABLED, NM_AVSTATE_NOT_INSTALLED);
-		oml_mo_state_chg(&trx->bb_transc.mo, -1, NM_AVSTATE_NOT_INSTALLED);
-
-		for (tn = 0; tn < ARRAY_SIZE(trx->ts); tn++)
-			oml_mo_state_chg(&trx->ts[tn].mo, NM_OPSTATE_DISABLED, NM_AVSTATE_NOT_INSTALLED);
-	}
-	return 0;
-}
-
 /* prepare the per-SAPI T200 arrays for a given lchan */
 static int t200_by_lchan(int *t200_ms_dcch, int *t200_ms_acch, struct gsm_lchan *lchan)
 {
@@ -398,6 +458,11 @@ static int t200_by_lchan(int *t200_ms_dcch, int *t200_ms_acch, struct gsm_lchan
 	t200_ms_acch[DL_SAPI0] = bts->t200_ms[T200_SACCH_SDCCH] + fn_advance_ms;
 	t200_ms_acch[DL_SAPI3] = bts->t200_ms[T200_SACCH_SDCCH] + fn_advance_ms;
 
+	if (lchan->repeated_acch_capability.dl_facch_all && (lchan->type == GSM_LCHAN_TCH_F || lchan->type == GSM_LCHAN_TCH_H)) {
+		t200_ms_acch[DL_SAPI0] *= 2;
+		t200_ms_acch[DL_SAPI3] *= 2;
+	}
+
 	switch (lchan->type) {
 	case GSM_LCHAN_SDCCH:
 		t200_ms_dcch[DL_SAPI0] = bts->t200_ms[T200_SDCCH] + fn_advance_ms;
@@ -428,7 +493,8 @@ int lchan_init_lapdm(struct gsm_lchan *lchan)
 			  "Setting T200 D0=%u, D3=%u, S0=%u, S3=%u (all in ms)\n",
 			  t200_ms_dcch[DL_SAPI0], t200_ms_dcch[DL_SAPI3],
 			  t200_ms_acch[DL_SAPI0], t200_ms_acch[DL_SAPI3]);
-		lapdm_channel_init2(lc, LAPDM_MODE_BTS, t200_ms_dcch, t200_ms_acch, lchan->type);
+		lapdm_channel_init3(lc, LAPDM_MODE_BTS, t200_ms_dcch, t200_ms_acch, lchan->type,
+				    gsm_lchan_name(lchan));
 		lapdm_channel_set_flags(lc, LAPDM_ENT_F_POLLING_ONLY);
 		lapdm_channel_set_l1(lc, NULL, lchan);
 	}
@@ -782,11 +848,6 @@ int bts_supports_cipher(struct gsm_bts *bts, int rsl_cipher)
 	return sup > 0;
 }
 
-bool trx_ms_pwr_ctrl_is_osmo(struct gsm_bts_trx *trx)
-{
-	return trx->ms_pwr_ctl_soft;
-}
-
 struct gsm_time *get_time(struct gsm_bts *bts)
 {
 	return &bts->gsm_time;
@@ -795,7 +856,7 @@ struct gsm_time *get_time(struct gsm_bts *bts)
 int bts_supports_cm(struct gsm_bts *bts, enum gsm_phys_chan_config pchan,
 		    enum gsm48_chan_mode cm)
 {
-	enum gsm_bts_features feature = _NUM_BTS_FEAT;
+	enum osmo_bts_features feature = _NUM_BTS_FEAT;
 
 	/* We assume that signalling support is mandatory,
 	 * there is no BTS_FEAT_* definition to check that. */
@@ -843,8 +904,29 @@ int bts_supports_cm(struct gsm_bts *bts, enum gsm_phys_chan_config pchan,
 	}
 
 	/* Check if the feature is supported by this BTS */
-	if (gsm_bts_has_feature(bts, feature))
+	if (osmo_bts_has_feature(bts->features, feature))
 		return 1;
 
 	return 0;
 }
+
+/* return the gsm_lchan for the CBCH (if it exists at all) */
+struct gsm_lchan *gsm_bts_get_cbch(struct gsm_bts *bts)
+{
+	struct gsm_lchan *lchan = NULL;
+	struct gsm_bts_trx *trx = bts->c0;
+
+	if (trx->ts[0].pchan == GSM_PCHAN_CCCH_SDCCH4_CBCH)
+		lchan = &trx->ts[0].lchan[2];
+	else {
+		int i;
+		for (i = 0; i < 8; i++) {
+			if (trx->ts[i].pchan == GSM_PCHAN_SDCCH8_SACCH8C_CBCH) {
+				lchan = &trx->ts[i].lchan[2];
+				break;
+			}
+		}
+	}
+
+	return lchan;
+}
